

#include "data_utility.h"


/*
*	Generate a random float value in [min_v, max_v].
*/
float rand_f( float min_v, float max_v)
{
	int rand_i;
	float rand_v;

	if( min_v > max_v)
		return 0;

//	srand( time(NULL));

	rand_i = rand( );

	rand_v = float( rand_i) / RAND_MAX;			//rand_v falls in [0,1] randomly.

	return (max_v - min_v) * rand_v + min_v;
}


/*
*	Generate a random integer from range [min_v, max_v].
*/	
int rand_i( int min_v, int max_v)
{
	if( max_v < min_v)
	{
		fprintf( stderr, "The range [%i, %i] is illegal.\n", min_v, max_v);
		return 0;
	}

	if( max_v == min_v)
		return min_v;

	if( max_v - min_v > RAND_MAX)
	{
		fprintf( stderr, "The random number to be generated is out-of-range(%i).\n", max_v - min_v);
		exit( 0);
	}

//	srand( time( NULL));

	return rand() % (max_v - min_v + 1) + min_v;
}


/*
*	enhance_rand_i generates an integer between min_v and max_v.
*	(max_v is 1000 times of RAND_MAX)
*
int enhance_rand_i( int min_v, int max_v)	
{
	int idx, offset, rand_i;

	if( min_v > max_v)
	{
		fprintf( stderr, "The range [%i, %i] is illegal.\n", min_v, max_v);
		return 0;
	}
	else if( min_v == max_v)
		return min_v;
	if( max_v - min_v > ENHANCE_RAND_TIMES * RAND_MAX)
	{
		fprintf( stderr, "The random number to be generated is out-of-range(%i).\n", max_v - min_v);
		exit( 0);
	}
	
	idx = rand()%ENHANCE_RAND_TIMES;
	offset = rand();
	
	rand_i = ENHANCE_RAND_TIMES * idx + offset;

	return rand_i % (max_v - min_v + 1) + min_v;
}
*/

void release_graph_copy( graph_t* graph_v)
{
	int i;

	g_node_t* g_node_v1, *g_node_v2;
	
	for( i=0; i<graph_v->size; i++)
	{
		g_node_v1 = graph_v->head_v[i].head;
		while( g_node_v1->next != NULL)
		{
			g_node_v2 = g_node_v1;
			g_node_v1 = g_node_v1->next;
			free( g_node_v2);
		}

		free( g_node_v1);
	}

	free( graph_v->head_v);
	free( graph_v);
}



config_t* read_config( )
{
	config_t* cfg;

	FILE* c_fp;

	cfg = (config_t*)malloc( sizeof( config_t));

	if( (c_fp = fopen( CONFIG_FILE, "r")) == NULL)
	{
		fprintf( stderr, "Cannot open the config file.\n");
		exit(0);
	}

	//Read the configuration information.
	fscanf( c_fp, "%s", cfg->g_file);
	fscanf( c_fp, "%i%i", &cfg->node_n, &cfg->edge_n);

	fscanf( c_fp, "%i%i", &cfg->J, &cfg->sample_n);

	fscanf( c_fp, "%lf", &cfg->ratio);

	fscanf( c_fp, "%i%i", &cfg->alg, &cfg->model);

	//Release related resources.
	fclose( c_fp);

	return cfg;
}

/*
*	Preprocess the power-law data generated by the tool "Python Web Graph Generator 2.72".
*	(http://mac.softpedia.com/get/Developer-Tools/Python-Web-Graph-Generator.shtml)
*/
void power_law_preprocess( )
{
	int i, edge_cnt, cnt = 0, node_n, id1, id2;
	char i_file[MAX_FILENAME_LENG];
	char o_file[MAX_FILENAME_LENG];

	char des[MAX_DESCRIPTION_LENG];

	FILE* c_fp, *i_fp, *o_fp;

	if( (c_fp = fopen( POWER_LAW_CONFIG, "r")) == NULL)
	{
		fprintf( stderr, "Cannot open the %s file.\n", POWER_LAW_CONFIG);
		exit( 0);
	}

	while( fscanf( c_fp, "%s%i", des, &node_n) != EOF)
	{
		printf( "Handling Case #%i.\n", ++cnt);

//		fscanf( c_fp, "%s%i", des, &edge_n);
		fscanf( c_fp, "%s%s", des, i_file);
		fscanf( c_fp, "%s%s", des, o_file);

		//Open the input & output files.
		if( ( i_fp = fopen( i_file, "r")) == NULL ||
			( o_fp = fopen( o_file, "w")) == NULL)
		{
			fprintf( stderr, "Cannot open i_file/o_file for Case #%i.", cnt);
			exit( 0);
		}

		//Collect the number of edges first.
		edge_cnt = 0;
		while( fscanf( i_fp, "%i%s%i", &id1, des, &id2) != EOF)
			edge_cnt ++;
		if( fseek( i_fp, 0, SEEK_SET) != 0)
		{
			fprintf( stderr, "fseek fails.\n");
			exit(0);
		}

		//Read & Write the data.
		fprintf( o_fp, "%i  %i\n", node_n, edge_cnt);
		for( i=0; i<edge_cnt; i++)
		{
			fscanf( i_fp, "%i%s%i", &id1, des, &id2);
			fprintf( o_fp, "%i  %i\n", id1-1, id2-1);
		}

		//Close the files.
		fclose( i_fp);
		fclose( o_fp);
	}

	//Release the related resources.
	fclose( c_fp);
}


/*
*	Remove th multiple edges in a graph.
*/
void remove_multiple_edges( )
{
	int i, node_n, edge_n, edge_n_cnt, idx1, idx2;
	
	char i_file[MAX_FILENAME_LENG];
	char o_file[MAX_FILENAME_LENG];
	char des[MAX_DESCRIPTION_LENG];

	FILE* c_fp, *i_fp, *o_fp;

	graph_t* graph_v;
	g_node_t* g_node_v, *g_node_v1;

	if( ( c_fp = fopen( REMOVE_MUL_EDGES_CONFIG, "r")) == NULL)
	{
		fprintf( stderr, "Cannot open remove_multiple_edges_config.txt.\n");
		exit( 0);
	}
	
	graph_v = (graph_t*)malloc( sizeof( graph_t));

	//Preprocess
	while( fscanf( c_fp, "%s%s", des, i_file) != EOF)
	{
		fscanf( c_fp, "%s%s", des, o_file);

		//Open the input & output files.
		if( ( i_fp = fopen( i_file, "r")) == NULL ||
			( o_fp = fopen( o_file, "w")) == NULL)
		{
			fprintf( stderr, "Cannot open the input/output file.\n");
			exit( 0);
		}

		//Read & Preprocess the input data.
		fscanf( i_fp, "%i%i", &node_n, &edge_n);

		graph_v->head_v = (g_head_t*)malloc( node_n * sizeof( g_head_t));
		
		for( i=0; i<node_n; i++)
		{
			g_node_v = (g_node_t*)malloc( sizeof( g_node_t));
			g_node_v->next = NULL;
			graph_v->head_v[i].head = g_node_v;
		}

		edge_n_cnt = 0;
		for( i=0; i<edge_n; i++)
		{
			fscanf( i_fp, "%i%i", &idx1, &idx2);

			g_node_v = graph_v->head_v[idx1].head->next;
			while( g_node_v != NULL && g_node_v->id != idx2)
				g_node_v = g_node_v->next;
			if( g_node_v == NULL)
			{
				g_node_v = (g_node_t*)malloc( sizeof( g_node_t));
				g_node_v->id = idx2;
				g_node_v->next = graph_v->head_v[idx1].head->next;
				graph_v->head_v[idx1].head->next = g_node_v;

				edge_n_cnt ++;
			}
		}

		//Output & Release the constructed graph without any multiple edges.
		fprintf( o_fp, "%i  %i\n", node_n, edge_n_cnt);
		for( i=0; i<node_n; i++)
		{
			g_node_v = graph_v->head_v[i].head->next;
			free( graph_v->head_v[i].head);
			while( g_node_v != NULL)
			{
				fprintf( o_fp, "%i  %i\n", i, g_node_v->id);

				g_node_v1 = g_node_v;
				g_node_v = g_node_v->next;
				free( g_node_v1);				
			}
		}
		free( graph_v->head_v);

		//Close the input & output files.
		fclose( i_fp);
		fclose( o_fp);
	}//

	//Release the malloced memory.
	free( graph_v);
	fclose( c_fp);
}


void sample_graph( )
{
	int i, node_n, edge_n, expect_node_n, edge_n_cnt, expect_tag, idx1, idx2, sta_n, sta, rear, cur_idx, chld_idx, c_cnt = 0;
	float sample_r;

	int *tra_tag;
	int *n_que;

	char i_file[ MAX_FILENAME_LENG];
	char o_file[ MAX_FILENAME_LENG];
	char des[ MAX_FILENAME_LENG];

	FILE* c_fp, *i_fp, *o_fp;

	graph_t* graph_v, *graph_v1;
	g_node_t* g_node_v, *g_node_v1;

	if( ( c_fp = fopen( SAMPLE_GRAPH_CONFIG, "r")) == NULL)
	{
		fprintf( stderr, "Cannot open the sample_graph_config file.\n");
		exit( 0);
	}

	while( fscanf( c_fp, "%s%i", des, &node_n) != EOF)
	{
		printf( "Handling case #%i...\n", ++c_cnt);

		fscanf( c_fp, "%s%f", des, &sample_r);
		fscanf( c_fp, "%s%s", des, i_file);
		fscanf( c_fp, "%s%s", des, o_file);

		if( ( i_fp = fopen( i_file, "r")) == NULL ||
			( o_fp = fopen( o_file, "w")) == NULL)
		{
			fprintf( stderr, "Cannot open the input/output files.\n");
			exit( 0);
		}

		//Calculate the expected the nodes.
		expect_node_n = int(node_n * sample_r);

		//Read the input file & Construct the corresponding graph structure.
		graph_v = (graph_t*)malloc( sizeof( graph_t));
		graph_v->head_v = (g_head_t*)malloc( node_n * sizeof( g_head_t));
		for( i=0; i<node_n; i++)
		{
			g_node_v = (g_node_t*)malloc( sizeof( g_node_t));
			memset( g_node_v, 0, sizeof( g_node_t));
			graph_v->head_v[i].head = g_node_v;
		}

		fscanf( i_fp, "%i%i", &node_n, &edge_n);
		while( fscanf( i_fp, "%i%i", &idx1, &idx2) != EOF)
		{
			g_node_v = (g_node_t*)malloc( sizeof( g_node_t));
			g_node_v->id = idx2;
			g_node_v->next = graph_v->head_v[idx1].head->next;
			graph_v->head_v[idx1].head->next = g_node_v;	
		}

		//Sample the graph via BFS until expect_node_n nodes have been traversed.
		graph_v1 = (graph_t*)malloc( sizeof( graph_t));
		graph_v1->head_v = (g_head_t*)malloc( expect_node_n * sizeof( g_head_t));
		for( i=0; i<expect_node_n; i++)
		{
			g_node_v = (g_node_t*)malloc( sizeof( g_node_t));
			g_node_v->next = NULL;
			graph_v1->head_v[i].head = g_node_v;
		}

		tra_tag = (int*)malloc( node_n * sizeof( int));
		for( i=0; i<node_n; i++)
			tra_tag[i] = -1;			//tag.

		n_que = (int*)malloc( node_n * sizeof( int));
		memset( n_que, 0, node_n * sizeof( int));

		sta = rear = 0;
		expect_tag = 0;
		edge_n_cnt = 0;
		
		while( rear < expect_node_n)
		{
			//Pick a start node.
			while( tra_tag[ sta_n = rand_i( 0, node_n-1)] != -1);	//enhance_rand_i( 0, node_n-1)

			n_que[ rear++] = sta_n;
			tra_tag[ sta_n] = rear - 1;

			if( rear >= expect_node_n)
			{
				//No neighbors.
				break;
			}

			//BFS.
			while( sta != rear)
			{
				cur_idx = n_que[ sta];

				g_node_v = graph_v->head_v[ cur_idx].head->next;
				while( g_node_v != NULL)
				{
					chld_idx = g_node_v->id;

					if( tra_tag[ chld_idx] == -1 && expect_tag == 0)
					{
						n_que[ rear++] = chld_idx;
						tra_tag[ chld_idx] = rear - 1;

						if( rear >= expect_node_n)
						{
							expect_tag = 1;
							//break;
						}
					}

					if( tra_tag[ chld_idx] != -1)
					{
						g_node_v1 = (g_node_t*)malloc( sizeof( g_node_t));
						g_node_v1->id = tra_tag[ chld_idx];
						g_node_v1->next = graph_v1->head_v[ sta].head->next;
						graph_v1->head_v[ sta].head->next = g_node_v1;

						edge_n_cnt ++;
					}

					g_node_v = g_node_v->next;
				}

				sta++;
			}//while( BFS).
		}//while( expect)

		//Output the sampled graph.
		fprintf( o_fp, "%i  %i\n", expect_node_n, edge_n_cnt);
/*t*/
		int temp = 0;
/*t*/
		for( i=0; i<expect_node_n; i++)
		{
			g_node_v1 = graph_v1->head_v[ i].head->next;
			while( g_node_v1 != NULL)
			{
				fprintf( o_fp, "%i  %i\n", i, g_node_v1->id);
				temp ++;

				g_node_v1 = g_node_v1->next;
			}
		}
/*t*/
		if( temp != edge_n_cnt)
		{
			fprintf( stderr, "Inconsistency of edge number.\n");
		}
/*t*/

		//Release the related resource.
		release_graph_copy( graph_v);
		release_graph_copy( graph_v1);
		free( tra_tag);
		free( n_que);
		fclose( i_fp);
		fclose( o_fp);
	
	}//while( config)

	//Release the related resource.

	fclose( c_fp);
}
